# -*- coding: utf-8 -*-
'''

'''
from os.path import abspath, exists, join, split, splitext

import cv2
import numpy as np

TILE_WIDTH = 256
TILE_HEIGHT = 256
TILE_SIZE = (TILE_WIDTH, TILE_HEIGHT)
TILE_OVERLAP = 64
TILE_OFFSET = 0
TILE_BORDERS = True


def compute(img_filepath, grid1=True, grid2=True, ext=None, **kwargs):
    """
    Compute the tiles for a given input image and saves them to disk.

    If a given tile has already been rendered to disk, it will not be recomputed.

    Args:
        img_filepath (str): image filepath (relative or absolute) to compute tiles for.
        grid1 (bool, optional): If :obj:`True`, create a dense grid of tiles on the image.
            Defaults to :obj:`True`.
        grid2 (bool, optional): If :obj:`True`, create a secondary dense grid of tiles
            on the image with a 50% offset.  Defaults to :obj:`True`.
        ext (str, optional): The file extension of the resulting tile files.  If this value is
            not specified, it will use the same extension as `img_filepath`.  Defaults
            to :obj:`None`.
        **kwargs: keyword arguments passed to :meth:`scoutbot.tile.tile_grid`

    Returns:
        tuple ( tuple ( int ), list ( dict ), list ( str ) ):
            - the original image's shape as ``(h, w, c)``.
            - list of grid coordinates as the output of :meth:`scoutbot.tile.tile_grid`.
            - list of tile filepaths as the output of :meth:`scoutbot.tile.tile_filepath`.
    """
    assert exists(img_filepath)
    img = cv2.imread(img_filepath)
    shape = img.shape

    grids = []
    if grid1:
        grids += tile_grid(shape, **kwargs)
    if grid2:
        grids += tile_grid(shape, offset=TILE_WIDTH // 2, borders=False, **kwargs)

    filepaths = [tile_filepath(img_filepath, grid, ext=ext) for grid in grids]
    for grid, filepath in zip(grids, filepaths):
        assert tile_write(img, grid, filepath)

    return shape, grids, filepaths


def tile_write(img, grid, filepath):
    """
    Write a single image's tile to disk using its grid coordinates and an output path.

    Args:
        img (numpy.ndarray): 3-dimentional Numpy array, the return from :func:`cv2.imread`
        grid (dict): the grid coordinate dictionary, one of the returned dictionaries
            from :meth:`scoutbot.tile.tile_grid`
        filepath (str): the tile's full output filepath (relative or absolute)

    Returns:
        bool: returns :obj:`True` if the tile's filepath exists on disk.
    """
    if exists(filepath):
        return True

    x0 = grid.get('x')
    y0 = grid.get('y')
    w = grid.get('w')
    h = grid.get('h')
    y1 = y0 + h
    x1 = x0 + w

    tile = img[y0:y1, x0:x1]
    cv2.imwrite(filepath, tile)
    return exists(filepath)


def tile_filepath(img_filepath, grid, ext=None):
    """
    Returns a suggested filepath for a tile given the original image filepath and the tile's grid
    coordinates.
    """
    x = grid.get('x')
    y = grid.get('y')
    w = grid.get('w')
    h = grid.get('h')

    assert exists(img_filepath)
    img_filepath = abspath(img_filepath)

    img_path, img_filename = split(img_filepath)
    img_name, img_ext = splitext(img_filename)

    img_ext = img_ext if ext is None else ext

    filepath = join(img_path, f'{img_name}_x_{x}_y_{y}_w_{w}_h_{h}{img_ext}')
    return filepath


def tile_grid(
    shape, size=TILE_SIZE, overlap=TILE_OVERLAP, offset=TILE_OFFSET, borders=TILE_BORDERS
):
    """
    Calculates a grid of tile coordinates for a given image.
    """
    h_, w_ = shape[:2]
    w, h = size
    ol = overlap
    os = offset

    if borders:
        assert offset == 0, 'Cannot use an offset with borders turned on'

    y_ = int(np.floor((h_ - ol) / (h - ol)))
    x_ = int(np.floor((w_ - ol) / (w - ol)))
    iy = (h * y_) - (ol * (y_ - 1))
    ix = (w * x_) - (ol * (x_ - 1))
    oy = int(np.floor((h_ - iy) * 0.5))
    ox = int(np.floor((w_ - ix) * 0.5))

    miny = 0
    minx = 0
    maxy = h_ - h
    maxx = w_ - w

    ys = list(range(oy, h_ - h + 1, h - ol))
    yb = [False] * len(ys)
    xs = list(range(ox, w_ - w + 1, w - ol))
    xb = [False] * len(xs)

    if borders and oy > 0:
        ys = [miny] + ys + [maxy]
        yb = [True] + yb + [True]

    if borders and ox > 0:
        xs = [minx] + xs + [maxx]
        xb = [True] + xb + [True]

    outputs = []
    for y0, yb_ in zip(ys, yb):
        y0 += os
        y1 = y0 + h
        for x0, xb_ in zip(xs, xb):
            x0 += os
            x1 = x0 + w

            # Sanity, mostly to check for offset
            valid = True
            try:
                assert x1 - x0 == w, '%d, %d' % (
                    x1 - x0,
                    w,
                )
                assert y1 - y0 == h, '%d, %d' % (
                    y1 - y0,
                    h,
                )
                assert 0 <= x0 and x0 <= w_, '%d, %d' % (
                    x0,
                    w_,
                )
                assert 0 <= x1 and x1 <= w_, '%d, %d' % (
                    x1,
                    w_,
                )
                assert 0 <= y0 and y0 <= h_, '%d, %d' % (
                    y0,
                    h_,
                )
                assert 0 <= y1 and y1 <= h_, '%d, %d' % (
                    y1,
                    h_,
                )
            except AssertionError:
                valid = False

            if valid:
                outputs.append({'x': x0, 'y': y0, 'w': w, 'h': h, 'b': yb_ or xb_})

    return outputs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Coming Soon
"""
import click

from scoutbot import utils

log = utils.init_logging()


@click.command()
@click.option(
    '--config', help='Path to config file', default='configs/mnist_resnet18.yaml'
)
def wic(config):
    """ """
    pass


@click.command()
@click.option(
    '--config', help='Path to config file', default='configs/mnist_resnet18.yaml'
)
def main(config):
    """ """
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Aggregation (AGG) returns unified detects for an image given its individual tile detections

This module defines how the tile-base localization detection results are aggregated
at the image level.  This includes the ability to weight the importance of detections
along the border of each tile within an image, and performing non-maximum suppression (NMS)
on the combined results.
"""
import numpy as np
import utool as ut

MARGIN = 32.0
AGG_THRESH = 0.4
NMS_THRESH = 0.2


def iou(box1, box2):
    """
    Computes the IoU (Intersection over Union) ratio for two bounding boxes.
    """
    inter_xtl = max(box1['xtl'], box2['xtl'])
    inter_ytl = max(box1['ytl'], box2['ytl'])
    inter_xbr = min(box1['xbr'], box2['xbr'])
    inter_ybr = min(box1['ybr'], box2['ybr'])

    inter_w = inter_xbr - inter_xtl
    inter_h = inter_ybr - inter_ytl

    if inter_w <= 0 or inter_h <= 0:
        inter = 0.0
    else:
        inter_w = max(0.0, inter_xbr - inter_xtl)
        inter_h = max(0.0, inter_ybr - inter_ytl)
        inter = inter_w * inter_h

    area1 = box1['w'] * box1['h']
    area2 = box2['w'] * box2['h']

    union = area1 + area2 - inter

    return area1, area2, inter, union


def demosaic(img_shape, tile_grids, loc_outputs, margin=MARGIN):
    """
    Demosaics a list of tiles and their respective detections back into the original
    image's coordinate system.
    """
    assert len(tile_grids) == len(loc_outputs)

    img_h, img_w = img_shape[:2]

    detects = []
    for tile_grid, loc_output in zip(tile_grids, loc_outputs):

        tile_xtl = tile_grid['x']
        tile_ytl = tile_grid['y']
        tile_w = tile_grid['w']
        tile_h = tile_grid['h']

        for detect in loc_output:
            detect_xtl = detect['x']
            detect_ytl = detect['y']
            detect_w = detect['w']
            detect_h = detect['h']
            detect_conf = detect['c']
            detect_label = detect['l']

            detect_xbr = detect_xtl + detect_w
            detect_ybr = detect_ytl + detect_h

            detect_box = {
                'xtl': detect_xtl / tile_w,
                'ytl': detect_ytl / tile_h,
                'xbr': detect_xbr / tile_w,
                'ybr': detect_ybr / tile_h,
                'w': detect_w / tile_w,
                'h': detect_h / tile_h,
            }

            margin_percent_w = margin / tile_w
            margin_percent_h = margin / tile_h

            center_box = {
                'xtl': margin_percent_w,
                'ytl': margin_percent_h,
                'xbr': 1.0 - margin_percent_w,
                'ybr': 1.0 - margin_percent_h,
                'w': 1.0 - (2.0 * margin_percent_w),
                'h': 1.0 - (2.0 * margin_percent_h),
            }
            area, _, inter, union = iou(detect_box, center_box)

            overlap = 0.0 if area <= 0 else inter / area
            overlap = round(overlap, 8)
            assert 0.0 <= overlap and overlap <= 1.0
            multiplier = np.sqrt(overlap)

            final_conf = round(detect_conf * multiplier, 4)
            if final_conf <= 0.0:
                continue

            final_xtl = int(np.around(tile_xtl + detect_xtl))
            final_ytl = int(np.around(tile_ytl + detect_ytl))
            final_w = int(np.around(detect_w))
            final_h = int(np.around(detect_h))
            final_xbr = final_xtl + final_w
            final_ybr = final_ytl + final_h

            # Check size with image frame
            final_xtl = min(max(final_xtl, 0), img_w)
            final_ytl = min(max(final_ytl, 0), img_h)
            final_xbr = min(max(final_xbr, 0), img_w)
            final_ybr = min(max(final_ybr, 0), img_h)
            final_w = final_xbr - final_xtl
            final_h = final_ybr - final_ytl

            final_area = final_w * final_h
            if final_area <= 0.0:
                continue

            detects.append(
                {
                    'l': detect_label,
                    'c': final_conf,
                    'x': final_xtl,
                    'y': final_ytl,
                    'w': final_w,
                    'h': final_h,
                }
            )

    return detects


def compute(
    img_shape, tile_grids, loc_outputs, agg_thresh=AGG_THRESH, nms_thresh=NMS_THRESH
):
    """
    Compute the aggregated image-level detection results for a given list of tile-level detections
    """
    from scoutbot.agg.py_cpu_nms import py_cpu_nms

    # Demosaic tile detection results and aggregate across the image
    detects = demosaic(img_shape, tile_grids, loc_outputs)

    # Filter low-confidence detections
    detects = [detect for detect in detects if detect['c'] >= agg_thresh]

    # Run NMS on aggregated detections
    coords = np.vstack(
        [
            [
                detect['x'],
                detect['y'],
                detect['x'] + detect['w'],
                detect['y'] + detect['h'],
            ]
            for detect in detects
        ]
    )
    confs = np.array([detect['c'] for detect in detects])

    keeps = py_cpu_nms(coords, confs, nms_thresh)
    final = ut.take(detects, keeps)
    final.sort(key=lambda val: val['c'], reverse=True)

    return final



# -*- coding: utf-8 -*-
import logging

log = logging.getLogger('pytest.conftest')  # pylint: disable=invalid-name


# @pytest.fixture()
# def cfg(config):
#     from scoutbot import utils

#     log = utils.init_logging()
#     cfg = utils.init_config(config, log)

#     cfg['output'] = 'scoutbot/{}'.format(cfg['output'])

#     return cfg


# @pytest.fixture()
# def device(cfg):
#     device = cfg.get('device')

#     return device


# @pytest.fixture()
# def net(cfg):
#     from scoutbot import model

#     net, _, _ = model.load(cfg)
#     net.eval()

#     return net



# -*- coding: utf-8 -*-
"""

pip install torch torchvision onnx onnxruntime-gpu tqdm wbia-utool scikit-learn numpy

"""
import random
import time
from collections import OrderedDict
from os.path import exists, join, split, splitext

import numpy as np
import onnx
import onnxruntime as ort
import sklearn
import torch
import torch.nn as nn
import torchvision
import tqdm
import utool as ut
import wbia
from wbia.algo.detect.densenet import INPUT_SIZE, ImageFilePathList, _init_transforms

WITH_GPU = False
BATCH_SIZE = 128


ibs = wbia.opendb(dbdir='/data/db')


pkl_path = 'scout.pkl'
if not exists(pkl_path):
    if False:
        tids = ibs.get_valid_gids(is_tile=True)
    else:
        imageset_text_list = ['TEST_SET']
        imageset_rowid_list = ibs.get_imageset_imgsetids_from_text(imageset_text_list)
        gids_list = ibs.get_imageset_gids(imageset_rowid_list)
        gids = ut.flatten(gids_list)
        flags = ibs.get_tile_flags(gids)
        test_gids = ut.filterfalse_items(gids, flags)
        assert sum(ibs.get_tile_flags(test_gids)) == 0
        tids = ibs.scout_get_valid_tile_rowids(gid_list=test_gids)

    random.shuffle(tids)
    positive, negative = [], []
    for chunk_tids in tqdm.tqdm(ut.ichunks(tids, 1000)):
        _, _, chunk_flags = ibs.scout_tile_positive_cumulative_area(chunk_tids)
        chunk_filepaths = ibs.get_image_paths(chunk_tids)
        for index, (tid, flag, filepath) in enumerate(
            zip(chunk_tids, chunk_flags, chunk_filepaths)
        ):
            if not exists(filepath):
                continue
            if flag:
                positive.append(tid)
            else:
                negative.append(tid)
        if len(positive) >= 100 and len(negative) >= 100:
            break
        print(len(positive), len(negative))

    random.shuffle(positive)
    random.shuffle(negative)
    positive = positive[:100]
    negative = negative[:100]
    data = positive + negative
    filepaths = ibs.get_image_paths(data)
    labels = [True] * len(positive) + [False] * len(negative)
    ut.save_cPkl(pkl_path, (data, labels))

    OUTPUT_PATH = '/data/db/checks'
    ut.delete(OUTPUT_PATH)
    ut.ensuredir(OUTPUT_PATH)
    for filepath, label in zip(filepaths, labels):
        path, filename = split(filepath)
        name, ext = splitext(filename)
        tag = 'true' if label else 'false'
        filename_ = f'{name}.{tag}{ext}'
        filepath_ = join(OUTPUT_PATH, filename_)
        if not exists(filepath_):
            ut.copy(filepath, filepath_)

assert exists(pkl_path)
data, labels = ut.load_cPkl(pkl_path)

filepaths = ibs.get_image_paths(data)

assert len(data) == len(set(data))
assert set(ibs.get_image_sizes(data)) == {(256, 256)}
assert sum(map(exists, filepaths)) == len(filepaths)

##########

INDEX = 0

weights_path = f'/cache/wbia/classifier2.scout.5fbfff26.3/classifier2.vulcan.5fbfff26.3/classifier.{INDEX}.weights'

assert exists(weights_path)
weights = torch.load(weights_path, map_location='cpu')
state = weights['state']
classes = weights['classes']

# Initialize the model for this run
model = torchvision.models.densenet201()
num_ftrs = model.classifier.in_features
model.classifier = nn.Linear(num_ftrs, len(classes))

# Convert any weights to non-parallel version
new_state = OrderedDict()
for k, v in state.items():
    k = k.replace('module.', '')
    new_state[k] = v

# Load state without parallel
model.load_state_dict(new_state)

# Add softmax
model.classifier = nn.Sequential(model.classifier, nn.LogSoftmax(), nn.Softmax())
if WITH_GPU:
    model = model.cuda()
model.eval()

#############

transforms = _init_transforms()
transform = transforms['test']
dataset = ImageFilePathList(filepaths, labels, transform=transform)
dataloader = torch.utils.data.DataLoader(
    dataset, batch_size=BATCH_SIZE, num_workers=0, pin_memory=False
)

time_pytorch = 0.0
inputs = []
outputs = []
targets = []
for (inputs_, targets_) in tqdm.tqdm(dataloader, desc='test'):
    if WITH_GPU:
        inputs_ = inputs_.cuda()

    time_start = time.time()
    with torch.set_grad_enabled(False):
        output_ = model(inputs_)
    time_end = time.time()
    time_pytorch += time_end - time_start

    inputs += inputs_.tolist()
    outputs += output_.tolist()
    targets += targets_.tolist()

inputs = np.array(inputs, dtype=np.float32)
globals().update(locals())
predictions_pytorch = [dict(zip(classes, output)) for output in outputs]

#############

threshs = list(np.arange(0.0, 1.01, 0.01))
best_thresh = None
best_accuracy = 0.0
best_confusion = None
for thresh in tqdm.tqdm(threshs):
    globals().update(locals())
    values = [prediction['positive'] >= thresh for prediction in predictions_pytorch]
    accuracy = sklearn.metrics.accuracy_score(targets, values)
    confusion = sklearn.metrics.confusion_matrix(targets, values)
    if accuracy > best_accuracy:
        best_thresh = thresh
        best_accuracy = accuracy
        best_confusion = confusion

tn, fp, fn, tp = best_confusion.ravel()
print(f'Thresh:    {best_thresh}')
print(f'Accuracy:  {best_accuracy}')
print(f'TP:        {tp}')
print(f'TN:        {tn}')
print(f'FP:        {fp}')
print(f'FN:        {fn}')

# Thresh:    0.01
# Accuracy:  0.895
# TP:        83
# TN:        96
# FP:        4
# FN:        17

# Thresh:    0.06
# Accuracy:  0.91
# TP:        85
# TN:        97
# FP:        3
# FN:        15

# Thresh:    0.01
# Accuracy:  0.905
# TP:        83
# TN:        98
# FP:        2
# FN:        17

#############

dummy_input = torch.randn(BATCH_SIZE, 3, INPUT_SIZE, INPUT_SIZE, device='cpu')
input_names = ['input']
output_names = ['output']

onnx_filename = f'scout.wic.5fbfff26.3.{INDEX}.onnx'
output = torch.onnx.export(
    model,
    dummy_input,
    onnx_filename,
    verbose=True,
    input_names=input_names,
    output_names=output_names,
    dynamic_axes={
        'input': {0: 'batch_size'},  # variable length axes
        'output': {0: 'batch_size'},
    },
)

###########

model = onnx.load(onnx_filename)
onnx.checker.check_model(model)
print(onnx.helper.printable_graph(model.graph))

###########

ort_session = ort.InferenceSession(onnx_filename, providers=['CPUExecutionProvider'])

time_onnx = 0.0
outputs = []
for chunk in ut.ichunks(inputs, BATCH_SIZE):
    trim = len(chunk)
    while (len(chunk)) < BATCH_SIZE:
        chunk.append(np.random.randn(3, INPUT_SIZE, INPUT_SIZE).astype(np.float32))
    input_ = np.array(chunk, dtype=np.float32)

    time_start = time.time()
    output_ = ort_session.run(
        None,
        {'input': input_},
    )
    time_end = time.time()
    time_onnx += time_end - time_start

    outputs += output_[0].tolist()[:trim]

predictions_onnx = [dict(zip(classes, output)) for output in outputs]

###########

values_pytorch = [
    prediction_pytorch['positive'] for prediction_pytorch in predictions_pytorch
]
values_onnx = [prediction_onnx['positive'] for prediction_onnx in predictions_onnx]
deviations = [
    abs(value_pytorch - value_onnx)
    for value_pytorch, value_onnx in zip(values_pytorch, values_onnx)
]

print(f'Min:  {np.min(deviations):0.08f}')
print(f'Max:  {np.max(deviations):0.08f}')
print(f'Mean: {np.mean(deviations):0.08f} +/- {np.std(deviations):0.08f}')
print(f'Time Pytorch: {time_pytorch:0.02f} sec.')
print(f'Time ONNX:    {time_onnx:0.02f} sec.')

globals().update(locals())
values = [prediction['positive'] >= best_thresh for prediction in predictions_onnx]
accuracy = sklearn.metrics.accuracy_score(targets, values)
confusion = sklearn.metrics.confusion_matrix(targets, values)
tn, fp, fn, tp = best_confusion.ravel()

print(f'Thresh:    {best_thresh}')
print(f'Accuracy:  {best_accuracy}')
print(f'TP:        {tp}')
print(f'TN:        {tn}')
print(f'FP:        {fp}')
print(f'FN:        {fn}')

# Min:  0.00000000
# Max:  0.00000143
# Mean: 0.00000003 +/- 0.00000013
# Time Pytorch: 9.64 sec.
# Time ONNX:    3.17 sec.
# Thresh:    0.01
# Accuracy:  0.895
# TP:        83
# TN:        96
# FP:        4
# FN:        17

# Min:  0.00000000
# Max:  0.00000113
# Mean: 0.00000004 +/- 0.00000013
# Time Pytorch: 9.42 sec.
# Time ONNX:    3.54 sec.
# Thresh:    0.06
# Accuracy:  0.91
# TP:        85
# TN:        97
# FP:        3
# FN:        15


# Min:  0.00000000
# Max:  0.00000209
# Mean: 0.00000004 +/- 0.00000019
# Time Pytorch: 9.98 sec.
# Time ONNX:    3.45 sec.
# Thresh:    0.01
# Accuracy:  0.905
# TP:        83
# TN:        98
# FP:        2
# FN:        17



# -*- coding: utf-8 -*-
"""
pip install torch torchvision onnx onnxruntime-gpu tqdm wbia-utool scikit-learn numpy
"""
import random
import time
from os.path import exists, join, split, splitext

import cv2
import lightnet as ln
import numpy as np
import onnx
import onnxruntime as ort
import sklearn
import torch
import torchvision
import tqdm
import utool as ut
import vtool as vt
import wbia

WITH_GPU = False
BATCH_SIZE = 16


ibs = wbia.opendb(dbdir='/data/db')


pkl_path = 'scout.pkl'
if not exists(pkl_path):
    if False:
        pass
        # tids = ibs.get_valid_gids(is_tile=True)
    else:
        imageset_text_list = ['TEST_SET']
        imageset_rowid_list = ibs.get_imageset_imgsetids_from_text(imageset_text_list)
        gids_list = ibs.get_imageset_gids(imageset_rowid_list)
        gids = ut.flatten(gids_list)
        flags = ibs.get_tile_flags(gids)
        test_gids = ut.filterfalse_items(gids, flags)
        assert sum(ibs.get_tile_flags(test_gids)) == 0
        tids = ibs.scout_get_valid_tile_rowids(gid_list=test_gids)

    random.shuffle(tids)
    positive, negative = [], []
    for chunk_tids in tqdm.tqdm(ut.ichunks(tids, 1000)):
        _, _, chunk_flags = ibs.scout_tile_positive_cumulative_area(chunk_tids)
        chunk_filepaths = ibs.get_image_paths(chunk_tids)
        for index, (tid, flag, filepath) in enumerate(
            zip(chunk_tids, chunk_flags, chunk_filepaths)
        ):
            if not exists(filepath):
                continue
            if flag:
                positive.append(tid)
            else:
                negative.append(tid)
        if len(positive) >= 100 and len(negative) >= 100:
            break
        print(len(positive), len(negative))

    random.shuffle(positive)
    random.shuffle(negative)
    positive = positive[:100]
    negative = negative[:100]
    data = positive + negative
    filepaths = ibs.get_image_paths(data)
    labels = [True] * len(positive) + [False] * len(negative)
    ut.save_cPkl(pkl_path, (data, labels))

    OUTPUT_PATH = '/data/db/checks'
    ut.delete(OUTPUT_PATH)
    ut.ensuredir(OUTPUT_PATH)
    for filepath, label in zip(filepaths, labels):
        path, filename = split(filepath)
        name, ext = splitext(filename)
        tag = 'true' if label else 'false'
        filename_ = f'{name}.{tag}{ext}'
        filepath_ = join(OUTPUT_PATH, filename_)
        if not exists(filepath_):
            ut.copy(filepath, filepath_)

assert exists(pkl_path)
data, labels = ut.load_cPkl(pkl_path)

filepaths = ibs.get_image_paths(data)
orients = ibs.get_image_orientation(data)

assert len(data) == len(set(data))
assert set(ibs.get_image_sizes(data)) == {(256, 256)}
assert sum(map(exists, filepaths)) == len(filepaths)
assert sum(orients) == 0

##########

INDEX = 1

config_path = f'/cache/lightnet/detect.lightnet.scout.5fbfff26.v{INDEX}.py'
weights_path = f'/cache/lightnet/detect.lightnet.scout.5fbfff26.v{INDEX}.weights'
conf_thresh = 0.0
nms_thresh = 0.2

assert exists(config_path)
assert exists(weights_path)

params = ln.engine.HyperParameters.from_file(config_path)
params.load(weights_path)

model = params.network

# Update conf_thresh and nms_thresh in postpsocess
model.postprocess[0].conf_thresh = conf_thresh
model.postprocess[1].nms_thresh = nms_thresh

if WITH_GPU:
    model = model.cuda()
model.eval()

INPUT_SIZE = params.input_dimension
INPUT_SIZE_H, INPUT_SIZE_W = INPUT_SIZE

#############

dataloader = list(zip(filepaths, orients, labels))

transform = torchvision.transforms.ToTensor()

time_pytorch = 0.0
inputs = []
sizes = []
outputs = []
targets = []
for chunk in ut.ichunks(dataloader, BATCH_SIZE):

    filepaths_ = ut.take_column(chunk, 0)
    orients_ = ut.take_column(chunk, 1)
    targets_ = ut.take_column(chunk, 2)

    inputs_ = []
    sizes_ = []
    for filepath, orient in zip(filepaths_, orients_):
        img = vt.imread(filepath, orient=orient)
        size = img.shape[:2][::-1]

        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = ln.data.transform.Letterbox.apply(img, dimension=INPUT_SIZE)
        img = transform(img)

        inputs_.append(img)
        sizes_.append(size)
    inputs_ = torch.stack(inputs_)

    if WITH_GPU:
        inputs_ = inputs_.cuda()

    time_start = time.time()
    with torch.set_grad_enabled(False):
        output_ = model(inputs_)
    time_end = time.time()
    time_pytorch += time_end - time_start

    output_transform_ = []
    for out_, size_ in zip(output_, sizes_):
        out_transform_ = ln.data.transform.ReverseLetterbox.apply(
            [out_], INPUT_SIZE, size_
        )
        output_transform_.append(out_transform_[0])

    inputs += inputs_.tolist()
    sizes += sizes_
    outputs += output_transform_
    targets += targets_

predictions_pytorch = outputs

#############

threshs = list(np.arange(0.0, 1.01, 0.01))
best_thresh = None
best_accuracy = 0.0
best_confusion = None
for thresh in tqdm.tqdm(threshs):
    globals().update(locals())
    values = [
        [prediction for prediction in predictions if prediction.confidence >= thresh]
        for predictions in predictions_pytorch
    ]
    values = [len(value) > 0 for value in values]
    accuracy = sklearn.metrics.accuracy_score(targets, values)
    confusion = sklearn.metrics.confusion_matrix(targets, values)
    if accuracy > best_accuracy:
        best_thresh = thresh
        best_accuracy = accuracy
        best_confusion = confusion

tn, fp, fn, tp = best_confusion.ravel()
print(f'Thresh:    {best_thresh}')
print(f'Accuracy:  {best_accuracy}')
print(f'TP:        {tp}')
print(f'TN:        {tn}')
print(f'FP:        {fp}')
print(f'FN:        {fn}')

# Thresh:    0.25
# Accuracy:  0.93
# TP:        88
# TN:        98
# FP:        2
# FN:        12

# Thresh:    0.35
# Accuracy:  0.925
# TP:        85
# TN:        100
# FP:        0
# FN:        15

#############

dummy_input = torch.randn(BATCH_SIZE, 3, INPUT_SIZE_H, INPUT_SIZE_W, device='cpu')
input_names = ['input']
output_names = ['output']

model.onnx = True
onnx_filename = f'scout.loc.5fbfff26.{INDEX}.onnx'
output = torch.onnx.export(
    model,
    dummy_input,
    onnx_filename,
    verbose=True,
    input_names=input_names,
    output_names=output_names,
    dynamic_axes={
        'input': {0: 'batch_size'},  # variable length axes
        'output': {0: 'batch_size'},
    },
)

###########

model = onnx.load(onnx_filename)
onnx.checker.check_model(model)
print(onnx.helper.printable_graph(model.graph))

###########

ort_session = ort.InferenceSession(onnx_filename, providers=['CPUExecutionProvider'])

num_classes = params.network.num_classes
anchors = params.network.anchors
network_size = (INPUT_SIZE_H, INPUT_SIZE_W, 3)
class_label_map = params.class_label_map
conf_thresh = 0.0
nms_thresh = 0.2

postprocess = ln.data.transform.Compose(
    [
        ln.data.transform.GetBoundingBoxes(num_classes, anchors, conf_thresh),
        ln.data.transform.NonMaxSupression(nms_thresh),
        ln.data.transform.TensorToBrambox(network_size, class_label_map),
    ]
)

zipped = list(zip(inputs, sizes))

time_onnx = 0.0
outputs = []
for chunk in ut.ichunks(zipped, BATCH_SIZE):

    imgs = ut.take_column(chunk, 0)
    sizes_ = ut.take_column(chunk, 1)

    trim = len(imgs)
    while (len(imgs)) < BATCH_SIZE:
        imgs.append(np.random.randn(3, INPUT_SIZE_H, INPUT_SIZE_W).astype(np.float32))
        sizes_.append(INPUT_SIZE)
    input_ = np.array(imgs, dtype=np.float32)

    time_start = time.time()
    outputs_ = ort_session.run(
        None,
        {'input': input_},
    )
    output_ = postprocess(torch.tensor(outputs_[0]))
    time_end = time.time()
    time_onnx += time_end - time_start

    output_transform_ = []
    for out_, size_ in zip(output_, sizes_):
        out_transform_ = ln.data.transform.ReverseLetterbox.apply(
            [out_], INPUT_SIZE, size_
        )
        output_transform_.append(out_transform_[0])

    outputs += output_transform_[:trim]

predictions_onnx = outputs

###########

globals().update(locals())
values_pytorch = [
    [prediction for prediction in predictions if prediction.confidence >= best_thresh]
    for predictions in predictions_pytorch
]
values_onnx = [
    [prediction for prediction in predictions if prediction.confidence >= best_thresh]
    for predictions in predictions_onnx
]

deviations = []
for value_pytorch, value_onnx in zip(values_pytorch, values_onnx):
    assert len(value_pytorch) == len(value_onnx)
    for value_p, value_o in zip(value_pytorch, value_onnx):
        assert value_p.class_label == value_o.class_label
        for attr in ['x_top_left', 'y_top_left', 'width', 'height', 'confidence']:
            deviation = abs(getattr(value_p, attr) - getattr(value_o, attr))
            deviations.append(deviation)

print(f'Min:  {np.min(deviations):0.08f}')
print(f'Max:  {np.max(deviations):0.08f}')
print(f'Mean: {np.mean(deviations):0.08f} +/- {np.std(deviations):0.08f}')
print(f'Time Pytorch: {time_pytorch:0.02f} sec.')
print(f'Time ONNX:    {time_onnx:0.02f} sec.')

values = [
    [prediction for prediction in predictions if prediction.confidence >= best_thresh]
    for predictions in predictions_onnx
]
values = [len(value) > 0 for value in values]
accuracy = sklearn.metrics.accuracy_score(targets, values)
confusion = sklearn.metrics.confusion_matrix(targets, values)
tn, fp, fn, tp = best_confusion.ravel()

print(f'Thresh:    {best_thresh}')
print(f'Accuracy:  {best_accuracy}')
print(f'TP:        {tp}')
print(f'TN:        {tn}')
print(f'FP:        {fp}')
print(f'FN:        {fn}')

# Min:  0.00000000
# Max:  0.00017841
# Mean: 0.00000904 +/- 0.00001550
# Time Pytorch: 18.18 sec.
# Time ONNX:    9.77 sec.
# Thresh:    0.25
# Accuracy:  0.93
# TP:        88
# TN:        98
# FP:        2
# FN:        12

# Min:  0.00000000
# Max:  0.00011268
# Mean: 0.00000845 +/- 0.00001284
# Time Pytorch: 18.75 sec.
# Time ONNX:    9.72 sec.
# Thresh:    0.35000000000000003
# Accuracy:  0.925
# TP:        85
# TN:        100
# FP:        0
# FN:        15



# -*- coding: utf-8 -*-
#
#   Lightnet related postprocessing
#   Thers are functions to transform the output of the network to brambox detection objects
#   Copyright EAVISE
#

import logging

import torch

# from torch.autograd import Variable
from scoutbot.loc.transforms.util import BaseTransform

__all__ = [
    'GetBoundingBoxes',
    'NonMaxSupression',
    'TensorToBrambox',
    'ReverseLetterbox',
]
log = logging.getLogger(__name__)


class GetBoundingBoxes(BaseTransform):
    """Convert output from darknet networks to bounding box tensor.

    Args:
        num_classes (int): number of categories
        anchors (list): 2D list representing anchor boxes (see :class:`lightnet.network.Darknet`)
        conf_thresh (Number [0-1]): Confidence threshold to filter detections

    Returns:
        (list [Batch x Tensor [Boxes x 6]]): **[x_center, y_center, width, height, confidence, class_id]** for every bounding box

    Note:
        The output tensor uses relative values for its coordinates.
    """

    def __init__(self, num_classes, anchors, conf_thresh):
        super().__init__(
            num_classes=num_classes, anchors=anchors, conf_thresh=conf_thresh
        )

    @classmethod
    def apply(cls, network_output, num_classes, anchors, conf_thresh):
        # Check dimensions
        if network_output.dim() == 3:
            network_output.unsqueeze_(0)

        # Variables
        num_anchors = len(anchors)
        # anchor_step = len(anchors[0])
        anchors = torch.Tensor(anchors)
        device = network_output.device
        batch = network_output.size(0)
        h = network_output.size(2)
        w = network_output.size(3)

        # Compute xc,yc, w,h, box_score on Tensor
        lin_x = torch.linspace(0, w - 1, w).repeat(h, 1).view(h * w).to(device)
        lin_y = torch.linspace(0, h - 1, h).view(h, 1).repeat(1, w).view(h * w).to(device)
        anchor_w = anchors[:, 0].contiguous().view(1, num_anchors, 1).to(device)
        anchor_h = anchors[:, 1].contiguous().view(1, num_anchors, 1).to(device)

        network_output = network_output.view(
            batch, num_anchors, -1, h * w
        )  # -1 == 5+num_classes (we can drop feature maps if 1 class)
        network_output[:, :, 0, :].sigmoid_().add_(lin_x).div_(w)  # X center
        network_output[:, :, 1, :].sigmoid_().add_(lin_y).div_(h)  # Y center
        network_output[:, :, 2, :].exp_().mul_(anchor_w).div_(w)  # Width
        network_output[:, :, 3, :].exp_().mul_(anchor_h).div_(h)  # Height
        network_output[:, :, 4, :].sigmoid_()  # Box score

        # Compute class_score
        if num_classes > 1:
            with torch.no_grad():
                cls_scores = torch.nn.functional.softmax(network_output[:, :, 5:, :], 2)
            cls_max, cls_max_idx = torch.max(cls_scores, 2)
            cls_max_idx = cls_max_idx.float()
            cls_max.mul_(network_output[:, :, 4, :])
        else:
            cls_max = network_output[:, :, 4, :]
            cls_max_idx = torch.zeros_like(cls_max)

        score_thresh = cls_max > conf_thresh
        score_thresh_flat = score_thresh.view(-1)

        if score_thresh.sum() == 0:
            boxes = []
            for i in range(batch):
                boxes.append(torch.tensor([]))
            return boxes

        # Mask select boxes > conf_thresh
        coords = network_output.transpose(2, 3)[..., 0:4]
        coords = coords[score_thresh[..., None].expand_as(coords)].view(-1, 4)
        scores = cls_max[score_thresh]
        idx = cls_max_idx[score_thresh]
        detections = torch.cat([coords, scores[:, None], idx[:, None]], dim=1)

        # Get indexes of splits between images of batch
        max_det_per_batch = num_anchors * h * w
        slices = [
            slice(max_det_per_batch * i, max_det_per_batch * (i + 1))
            for i in range(batch)
        ]
        det_per_batch = torch.IntTensor(
            [score_thresh_flat[s].int().sum() for s in slices]
        )
        split_idx = torch.cumsum(det_per_batch, dim=0)

        # Group detections per image of batch
        boxes = []
        start = 0
        for end in split_idx:
            boxes.append(detections[start:end])
            start = end

        return boxes


class NonMaxSupression(BaseTransform):
    """Performs nms on the bounding boxes, filtering boxes with a high overlap.

    Args:
        nms_thresh (Number [0-1]): Overlapping threshold to filter detections with non-maxima suppresion
        class_nms (Boolean, optional): Whether to perform nms per class; Default **True**

    Returns:
        (list [Batch x Tensor [Boxes x 6]]): **[x_center, y_center, width, height, confidence, class_id]** for every bounding box

    Note:
        This post-processing function expects the input to be bounding boxes,
        like the ones created by :class:`lightnet.data.GetBoundingBoxes` and outputs exactly the same format.
    """

    def __init__(self, nms_thresh, class_nms=True):
        super().__init__(nms_thresh=nms_thresh, class_nms=class_nms)

    @classmethod
    def apply(cls, boxes, nms_thresh, class_nms=True):
        return [cls._nms(box, nms_thresh, class_nms) for box in boxes]

    @staticmethod
    def _nms(boxes, nms_thresh, class_nms):
        """Non maximum suppression.

        Args:
          boxes (tensor): Bounding boxes of one image

        Return:
          (tensor): Pruned boxes
        """
        if boxes.numel() == 0:
            return boxes

        a = boxes[:, :2]
        b = boxes[:, 2:4]
        bboxes = torch.cat([a - b / 2, a + b / 2], 1)
        scores = boxes[:, 4]
        classes = boxes[:, 5]

        # Sort coordinates by descending score
        scores, order = scores.sort(0, descending=True)
        x1, y1, x2, y2 = bboxes[order].split(1, 1)

        # Compute dx and dy between each pair of boxes (these mat contain every pair twice...)
        dx = (x2.min(x2.t()) - x1.max(x1.t())).clamp(min=0)
        dy = (y2.min(y2.t()) - y1.max(y1.t())).clamp(min=0)

        # Compute iou
        intersections = dx * dy
        areas = (x2 - x1) * (y2 - y1)
        unions = (areas + areas.t()) - intersections
        ious = intersections / unions

        # Filter based on iou (and class)
        conflicting = (ious > nms_thresh).triu(1)

        if class_nms:
            classes = classes[order]
            same_class = classes.unsqueeze(0) == classes.unsqueeze(1)
            conflicting = conflicting & same_class

        conflicting = conflicting.cpu()
        keep = torch.zeros(len(conflicting), dtype=bool)
        supress = torch.zeros(len(conflicting), dtype=torch.float)
        for i, row in enumerate(conflicting):
            if not supress[i]:
                keep[i] = 1
                supress[row] = 1

        return boxes[order][keep[:, None].expand_as(boxes)].view(-1, 6).contiguous()


class TensorToBrambox(BaseTransform):
    """Converts a tensor to a list of brambox objects.

    Args:
        network_size (tuple): Tuple containing the width and height of the images going in the network
        class_label_map (list, optional): List of class labels to transform the class id's in actual names; Default **None**

    Returns:
        (list [list [brambox.boxes.Detection]]): list of brambox detections per image

    Note:
        If no `class_label_map` is given, this transform will simply convert the class id's in a string.

    Note:
        Just like everything in PyTorch, this transform only works on batches of images.
        This means you need to wrap your tensor of detections in a list if you want to run this transform on a single image.
    """

    def __init__(self, network_size, class_label_map=None):
        super().__init__(network_size=network_size, class_label_map=class_label_map)
        if self.class_label_map is None:
            log.warn(
                'No class_label_map given. The indexes will be used as class_labels.'
            )

    @classmethod
    def apply(cls, boxes, network_size, class_label_map=None):
        converted_boxes = []
        for box in boxes:
            if box.numel() == 0:
                converted_boxes.append([])
            else:
                converted_boxes.append(
                    cls._convert(box, network_size[0], network_size[1], class_label_map)
                )
        return converted_boxes

    @staticmethod
    def _convert(boxes, width, height, class_label_map):
        from scoutbot.loc.transforms.detections.detection import Detection

        boxes[:, 0:3:2].mul_(width)
        boxes[:, 0] -= boxes[:, 2] / 2
        boxes[:, 1:4:2].mul_(height)
        boxes[:, 1] -= boxes[:, 3] / 2

        brambox = []
        for box in boxes:
            det = Detection()
            det.x_top_left = box[0].item()
            det.y_top_left = box[1].item()
            det.width = box[2].item()
            det.height = box[3].item()
            det.confidence = box[4].item()
            if class_label_map is not None:
                det.class_label = class_label_map[int(box[5].item())]
            else:
                det.class_label = str(int(box[5].item()))

            brambox.append(det)

        return brambox


class ReverseLetterbox(BaseTransform):
    """Performs a reverse letterbox operation on the bounding boxes, so they can be visualised on the original image.

    Args:
        network_size (tuple): Tuple containing the width and height of the images going in the network
        image_size (tuple): Tuple containing the width and height of the original images

    Returns:
        (list [list [brambox.boxes.Detection]]): list of brambox detections per image

    Note:
        This transform works on :class:`brambox.boxes.Detection` objects,
        so you need to apply the :class:`~lightnet.data.TensorToBrambox` transform first.

    Note:
        Just like everything in PyTorch, this transform only works on batches of images.
        This means you need to wrap your tensor of detections in a list if you want to run this transform on a single image.
    """

    def __init__(self, network_size, image_size):
        super().__init__(network_size=network_size, image_size=image_size)

    @classmethod
    def apply(cls, boxes, network_size, image_size):
        im_w, im_h = image_size[:2]
        net_w, net_h = network_size[:2]

        if im_w == net_w and im_h == net_h:
            scale = 1
        elif im_w / net_w >= im_h / net_h:
            scale = im_w / net_w
        else:
            scale = im_h / net_h
        pad = int((net_w - im_w / scale) / 2), int((net_h - im_h / scale) / 2)

        converted_boxes = []
        for b in boxes:
            converted_boxes.append(cls._transform(b, scale, pad))
        return converted_boxes

    @staticmethod
    def _transform(boxes, scale, pad):
        for box in boxes:
            box.x_top_left -= pad[0]
            box.y_top_left -= pad[1]

            box.x_top_left *= scale
            box.y_top_left *= scale
            box.width *= scale
            box.height *= scale
        return boxes



# -*- coding: utf-8 -*-
'''The localizer (LOC) returns bounding box detections on image tiles.

This module defines how Localizer models are downloaded from an external CDN,
how to load an image and prepare it for inference, demonstrates how to run the
Localization ONNX model on this input, and finally how to convert this raw CNN
output into usable detection bounding boxes with class labels and confidence
scores.
'''
from os.path import exists, join
from pathlib import Path

import cv2
import numpy as np
import onnxruntime as ort
import pooch
import torch
import torchvision
import utool as ut

from scoutbot.loc.transforms import (
    Compose,
    GetBoundingBoxes,
    Letterbox,
    NonMaxSupression,
    ReverseLetterbox,
    TensorToBrambox,
)

PWD = Path(__file__).absolute().parent

BATCH_SIZE = 16
INPUT_SIZE = (416, 416)
INPUT_SIZE_H, INPUT_SIZE_W = INPUT_SIZE
NETWORK_SIZE = (INPUT_SIZE_H, INPUT_SIZE_W, 3)

NUM_CLASSES = 1
ANCHORS = [
    (1.3221, 1.73145),
    (3.19275, 4.00944),
    (5.05587, 8.09892),
    (9.47112, 4.84053),
    (11.2364, 10.0071),
]
CLASS_LABEL_MAP = ['elephant_savanna']
LOC_THRESH = 0.4
NMS_THRESH = 0.8

ONNX_MODEL = 'scout.loc.5fbfff26.0.onnx'
ONNX_MODEL_PATH = join(PWD, 'models', 'onnx', ONNX_MODEL)
ONNX_MODEL_HASH = '85a9378311d42b5143f74570136f32f50bf97c548135921b178b46ba7612b216'


def fetch(pull=False):
    """
    Fetch the Localizer ONNX model file from a CDN if it does not exist locally.

    This function will throw an AssertionError if the download fails or the
    file otherwise does not exists locally on disk.

    Args:
        pull (bool, optional): If :obj:`True`, use a downloaded version stored in
            the local system's cache.  Defaults to :obj:`False`.

    Returns:
        str: local ONNX model file path.

    Raises:
        AssertionError: If the model cannot be fetched.
    """
    if not pull and exists(ONNX_MODEL_PATH):
        onnx_model = ONNX_MODEL_PATH
    else:
        onnx_model = pooch.retrieve(
            url=f'https://wildbookiarepository.azureedge.net/models/{ONNX_MODEL}',
            known_hash=ONNX_MODEL_HASH,
            progressbar=True,
        )
        assert exists(onnx_model)

    return onnx_model


def pre(inputs):
    """
    Load a list of filepaths and return a corresponding list of the image
    data as a 4-D list of floats.  The image data is loaded from disk, transformed
    as needed, and is normalized to the input ranges that the Localizer ONNX model
    expects.

    This function will throw an error if any of the filepaths do not exist.

    Args:
        inputs (list(str)): list of tile image filepaths (relative or absolute)

    Returns:
        tuple ( list ( list ( list ( list ( float ) ) ) ), list ( tuple ( int ) ) ):
            - list of transformed image data.
            - list of each tile's original size.
    """
    transform = torchvision.transforms.ToTensor()

    data = []
    sizes = []
    for filepath in inputs:
        img = cv2.imread(filepath)
        size = img.shape[:2][::-1]

        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = Letterbox.apply(img, dimension=INPUT_SIZE)
        img = transform(img)

        data.append(img.tolist())
        sizes.append(size)

    return data, sizes


def predict(data, fill=True):
    """
    Run neural network inference using the Localizer's ONNX model on preprocessed data.

    Args:
        data (list): list of transformed image data, the first return of :meth:`scoutbot.loc.pre`
        fill (bool, optional): If :obj:`True`, fill any partial batches to the LOC `BATCH_SIZE`,
            and then trim them after inference.  Defaults to :obj:`True`.

    Returns:
        list ( list ( float ) ): list of raw ONNX model outputs
    """
    onnx_model = fetch()

    ort_session = ort.InferenceSession(
        onnx_model, providers=['CUDAExecutionProvider', 'CPUExecutionProvider']
    )

    preds = []
    for chunk in ut.ichunks(data, BATCH_SIZE):
        trim = len(chunk)
        if fill:
            while (len(chunk)) < BATCH_SIZE:
                chunk.append(
                    np.random.randn(3, INPUT_SIZE_H, INPUT_SIZE_W).astype(np.float32)
                )
        input_ = np.array(chunk, dtype=np.float32)

        pred_ = ort_session.run(
            None,
            {'input': input_},
        )
        preds += pred_[0].tolist()[:trim]

    return preds


def post(preds, sizes, loc_thresh=LOC_THRESH, nms_thresh=NMS_THRESH):
    """
    Apply a post-processing normalization of the raw ONNX network outputs.

    The final output is a list of lists of dictionaries, each representing a single
    detection.  Each dictionary has a structure with the following keys:

        ::

            {
                'l': class_label (str)
                'c': confidence (float)
                'x': x_top_left (float)
                'y': y_top_left (float)
                'w': width (float)
                'h': height (float)
            }

    The ``l`` label is the string class as used when the original
    ONNX model was trained.

    The ``c`` confidence value is a bounded float between ``0.0`` and
    ``1.0`` (inclusive), but should not be treated as a probability.

    The ``x``, ``y``, ``w``, ``h`` bounding box keys are in real pixel values.

    Args:
        preds (list): list of raw ONNX model outputs, the return of :meth:`scoutbot.loc.predict`
        sizes (list): list of original tile sizes, the second return of :meth:`scoutbot.loc.pre`

    Returns:
        list ( list ( dict ) ): nested list of Localizer predictions
    """
    postprocess = Compose(
        [
            GetBoundingBoxes(NUM_CLASSES, ANCHORS, loc_thresh),
            NonMaxSupression(nms_thresh),
            TensorToBrambox(NETWORK_SIZE, CLASS_LABEL_MAP),
        ]
    )

    preds = postprocess(torch.tensor(preds))

    outputs = []
    for pred, size in zip(preds, sizes):
        output = ReverseLetterbox.apply([pred], INPUT_SIZE, size)
        output = output[0]
        output = [
            {
                'l': detect.class_label,
                'c': detect.confidence,
                'x': detect.x_top_left,
                'y': detect.y_top_left,
                'w': detect.width,
                'h': detect.height,
            }
            for detect in output
        ]
        outputs.append(output)

    return outputs



# -*- coding: utf-8 -*-
#
#   Lightnet related data processing
#   Utilitary classes and functions for the data subpackage
#   Copyright EAVISE
#

from abc import ABC, abstractmethod

__all__ = ['Compose']


class Compose(list):
    """This is lightnet's own version of :class:`torchvision.transforms.Compose`.

    Note:
        The reason we have our own version is because this one offers more freedom to the user.
        For all intends and purposes this class is just a list.
        This `Compose` version allows the user to access elements through index, append items, extend it with another list, etc.
        When calling instances of this class, it behaves just like :class:`torchvision.transforms.Compose`.

    Note:
        I proposed to change :class:`torchvision.transforms.Compose` to something similar to this version,
        which would render this class useless. In the meanwhile, we use our own version
        and you can track `the issue`_ to see if and when this comes to torchvision.

    Ignore:
        >>> tf = ln.data.transform.Compose([lambda n: n+1])
        >>> tf(10)  # 10+1
        11
        >>> tf.append(lambda n: n*2)
        >>> tf(10)  # (10+1)*2
        22
        >>> tf.insert(0, lambda n: n//2)
        >>> tf(10)  # ((10//2)+1)*2
        12
        >>> del tf[2]
        >>> tf(10)  # (10//2)+1
        6

    .. _the issue: https://github.com/pytorch/vision/issues/456
    """

    def __call__(self, data):
        for tf in self:
            data = tf(data)
        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + ' ['
        for tf in self:
            format_string += '\n  {tf}'
        format_string += '\n]'
        return format_string


class BaseTransform(ABC):
    """Base transform class for the pre- and post-processing functions.
    This class allows to create an object with some case specific settings, and then call it with the data to perform the transformation.
    It also allows to call the static method ``apply`` with the data and settings. This is usefull if you want to transform a single data object.
    """

    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __call__(self, data):
        return self.apply(data, **self.__dict__)

    @classmethod
    @abstractmethod
    def apply(cls, data, **kwargs):
        """Classmethod that applies the transformation once.

        Args:
            data: Data to transform (eg. image)
            **kwargs: Same arguments that are passed to the ``__init__`` function
        """
        return data


class BaseMultiTransform(ABC):
    """Base multiple transform class that is mainly used in pre-processing functions.
    This class exists for transforms that affect both images and annotations.
    It provides a classmethod ``apply``, that will perform the transormation on one (data, target) pair.
    """

    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    @abstractmethod
    def __call__(self, data):
        return data

    @classmethod
    def apply(cls, data, target=None, **kwargs):
        """Classmethod that applies the transformation once.

        Args:
            data: Data to transform (eg. image)
            target (optional): ground truth for that data; Default **None**
            **kwargs: Same arguments that are passed to the ``__init__`` function
        """
        obj = cls(**kwargs)
        res_data = obj(data)

        if target is None:
            return res_data

        res_target = obj(target)
        return res_data, res_target



# -*- coding: utf-8 -*-
'''The Whole Image Classifier (WIC) returns confidence scores for image tiles.

This module defines how WIC models are downloaded from an external CDN,
how to load an image and prepare it for inference, demonstrates how to run the
WIC ONNX model on this input, and finally how to convert this raw CNN output
into usable confidence scores.
'''
from os.path import exists, join
from pathlib import Path

import numpy as np
import onnxruntime as ort
import pooch
import torch
import utool as ut

from scoutbot.wic.dataloader import (
    BATCH_SIZE,
    INPUT_SIZE,
    ImageFilePathList,
    _init_transforms,
)

PWD = Path(__file__).absolute().parent

ONNX_MODEL = 'scout.wic.5fbfff26.3.0.onnx'
ONNX_MODEL_PATH = join(PWD, 'models', 'onnx', ONNX_MODEL)
ONNX_MODEL_HASH = 'cbc7f381fa58504e03b6510245b6b2742d63049429337465d95663a6468df4c1'
ONNX_CLASSES = ['negative', 'positive']

WIC_THRESH = 0.2


def fetch(pull=False):
    """
    Fetch the WIC ONNX model file from a CDN if it does not exist locally.

    This function will throw an AssertionError if the download fails or the
    file otherwise does not exists locally on disk.

    Args:
        pull (bool, optional): If :obj:`True`, use a downloaded version stored in
            sthe local system's cache.  Defaults to :obj:`False`.

    Returns:
        str: local ONNX model file path.

    Raises:
        AssertionError: If the model cannot be fetched.
    """
    if not pull and exists(ONNX_MODEL_PATH):
        onnx_model = ONNX_MODEL_PATH
    else:
        onnx_model = pooch.retrieve(
            url=f'https://wildbookiarepository.azureedge.net/models/{ONNX_MODEL}',
            known_hash=ONNX_MODEL_HASH,
            progressbar=True,
        )
        assert exists(onnx_model)

    return onnx_model


def pre(inputs):
    """
    Load a list of filepaths and return a corresponding list of the image
    data as a 4-D list of floats.  The image data is loaded from disk, transformed
    as needed, and is normalized to the input ranges that the WIC ONNX model
    expects.

    This function will throw an error if any of the filepaths do not exist.

    Args:
        inputs (list(str)): list of tile image filepaths (relative or absolute)

    Returns:
        list ( list ( list ( list ( float ) ) ) ): list of transformed image data
    """
    transform = _init_transforms()
    dataset = ImageFilePathList(inputs, transform=transform)
    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=BATCH_SIZE, num_workers=0, pin_memory=False
    )

    data = []
    for (data_,) in dataloader:
        data += data_.tolist()

    return data


def predict(data, fill=False):
    """
    Run neural network inference using the WIC's ONNX model on preprocessed data.

    Args:
        data (list): list of transformed image data, the return of :meth:`scoutbot.wic.pre`
        fill (bool, optional): If :obj:`True`, fill any partial batches to the WIC `BATCH_SIZE`,
            and then trim them after inference.  Defaults to :obj:`False`.

    Returns:
        list ( list ( float ) ): list of raw ONNX model outputs
    """
    onnx_model = fetch()

    ort_session = ort.InferenceSession(
        onnx_model, providers=['CUDAExecutionProvider', 'CPUExecutionProvider']
    )

    preds = []
    for chunk in ut.ichunks(data, BATCH_SIZE):
        trim = len(chunk)
        if fill:
            while (len(chunk)) < BATCH_SIZE:
                chunk.append(
                    np.random.randn(3, INPUT_SIZE, INPUT_SIZE).astype(np.float32)
                )
        input_ = np.array(chunk, dtype=np.float32)

        pred_ = ort_session.run(
            None,
            {'input': input_},
        )
        preds += pred_[0].tolist()[:trim]

    return preds


def post(preds):
    """
    Apply a post-processing normalization of the raw ONNX network outputs.

    The final output is a dictionary where the key values are the predicted labels
    and the values are their corresponding confidence values.

    Args:
        preds (list): list of raw ONNX model outputs, the return of :meth:`scoutbot.wic.predict`

    Returns:
        list ( dict ): list of WIC predictions
    """
    outputs = [dict(zip(ONNX_CLASSES, pred)) for pred in preds]
    return outputs



# -*- coding: utf-8 -*-
#
#   Lightnet data transforms
#   Copyright EAVISE
#

from scoutbot.loc.transforms import annotations  # NOQA
from scoutbot.loc.transforms import box  # NOQA
from scoutbot.loc.transforms import detections  # NOQA
from scoutbot.loc.transforms import util  # NOQA
from scoutbot.loc.transforms._postprocess import (  # NOQA
    GetBoundingBoxes,
    NonMaxSupression,
    ReverseLetterbox,
    TensorToBrambox,
)
from scoutbot.loc.transforms._preprocess import Letterbox  # NOQA



# -*- coding: utf-8 -*-
'''
The above components must be run in the correct order, but ScoutbBot also offers a single pipeline.

All of the ML models can be pre-downloaded and fetched in a single call to :func:`scoutbot.fetch` and
the unified pipeline -- which uses the 4 components correctly -- can be run by the function
:func:`scoutbot.pipeline`.  Below is example code for how these components interact.

Furthermore, there are two application demo files (``app.py`` and ``app2.py``) that shows
how the entire pipeline can be run on tiles or images, respectively.

.. code-block:: python

    # Get image filepath
    filepath = '/path/to/image.ext'

    # Run tiling
    img_shape, tile_grids, tile_filepaths = tile.compute(filepath)

    # Run WIC
    wic_outputs = wic.post(wic.predict(wic.pre(tile_filepaths)))

    # Threshold for WIC
    flags = [wic_output.get('positive') >= wic_thresh for wic_output in wic_outputs]
    loc_tile_grids = ut.compress(tile_grids, flags)
    loc_tile_filepaths = ut.compress(tile_filepaths, flags)

    # Run localizer
    loc_data, loc_sizes = loc.pre(loc_tile_filepaths)
    loc_preds = loc.predict(loc_data)
    loc_outputs = loc.post(
        loc_preds,
        loc_sizes,
        loc_thresh=loc_thresh,
        nms_thresh=loc_nms_thresh
    )

    # Run Aggregation and get final detections
    detects = agg.compute(
        img_shape,
        loc_tile_grids,
        loc_outputs,
        agg_thresh=agg_thresh,
        nms_thresh=agg_nms_thresh,
    )
'''
from scoutbot import agg, loc, tile, wic

VERSION = '0.1.6'
version = VERSION
__version__ = VERSION


def fetch(pull=False):
    """
    Fetch the WIC and Localizer ONNX model files from a CDN if they do not exist locally.

    This function will throw an AssertionError if either download fails or the
    files otherwise do not exist locally on disk.

    Args:
        pull (bool, optional): If :obj:`True`, use the downloaded versions stored in
            the local system's cache.  Defaults to :obj:`False`.

    Returns:
        None

    Raises:
        AssertionError: If any model cannot be fetched.
    """
    wic.fetch(pull=pull)
    loc.fetch(pull=pull)


def pipeline(
    filepath,
    wic_thresh=wic.WIC_THRESH,
    loc_thresh=loc.LOC_THRESH,
    loc_nms_thresh=loc.NMS_THRESH,
    agg_thresh=agg.AGG_THRESH,
    agg_nms_thresh=agg.NMS_THRESH,
):
    """
    Run the ML pipeline on a given image filepath and return the detections

    The final output is a list of dictionaries, each representing a single detection.
    Each dictionary has a structure with the following keys:

        ::

            {
                'l': class_label (str)
                'c': confidence (float)
                'x': x_top_left (float)
                'y': y_top_left (float)
                'w': width (float)
                'h': height (float)
            }

    Args:
        filepath (str): image filepath (relative or absolute)

    Returns:
        list ( dict ): list of predictions
    """
    import utool as ut

    # Run tiling
    img_shape, tile_grids, tile_filepaths = tile.compute(filepath)

    # Run WIC
    wic_outputs = wic.post(wic.predict(wic.pre(tile_filepaths)))

    # Threshold for WIC
    flags = [wic_output.get('positive') >= wic_thresh for wic_output in wic_outputs]
    loc_tile_grids = ut.compress(tile_grids, flags)
    loc_tile_filepaths = ut.compress(tile_filepaths, flags)

    # Run localizer
    loc_data, loc_sizes = loc.pre(loc_tile_filepaths)
    loc_preds = loc.predict(loc_data)
    loc_outputs = loc.post(
        loc_preds, loc_sizes, loc_thresh=loc_thresh, nms_thresh=loc_nms_thresh
    )
    assert len(loc_tile_grids) == len(loc_outputs)

    # Run Aggregation
    detects = agg.compute(
        img_shape,
        loc_tile_grids,
        loc_outputs,
        agg_thresh=agg_thresh,
        nms_thresh=agg_nms_thresh,
    )

    return detects



# -*- coding: utf-8 -*-
from os.path import abspath, join

import utool as ut

from scoutbot import agg, loc, tile, wic


def test_agg_compute():
    img_filepath = abspath(join('examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41.jpg'))

    # Run tiling
    img_shape, tile_grids, tile_filepaths = tile.compute(img_filepath)
    assert len(tile_filepaths) == 1252

    # Run WIC
    wic_outputs = wic.post(wic.predict(wic.pre(tile_filepaths)))
    assert len(wic_outputs) == len(tile_filepaths)

    # Threshold for WIC
    flags = [wic_output.get('positive') >= wic.WIC_THRESH for wic_output in wic_outputs]
    loc_tile_grids = ut.compress(tile_grids, flags)
    loc_tile_filepaths = ut.compress(tile_filepaths, flags)
    assert sum(flags) == 15

    # Run localizer
    loc_data, loc_sizes = loc.pre(loc_tile_filepaths)
    loc_preds = loc.predict(loc_data)
    loc_outputs = loc.post(
        loc_preds, loc_sizes, loc_thresh=loc.LOC_THRESH, nms_thresh=loc.NMS_THRESH
    )
    assert len(loc_tile_grids) == len(loc_outputs)

    # Aggregate
    detects = agg.compute(
        img_shape,
        loc_tile_grids,
        loc_outputs,
        agg_thresh=agg.AGG_THRESH,
        nms_thresh=agg.NMS_THRESH,
    )

    assert len(detects) == 3

    targets = [
        {'l': 'elephant_savanna', 'c': 0.9299, 'x': 4597, 'y': 2322, 'w': 72, 'h': 149},
        {'l': 'elephant_savanna', 'c': 0.8739, 'x': 4865, 'y': 2422, 'w': 97, 'h': 109},
        {'l': 'elephant_savanna', 'c': 0.7115, 'x': 4806, 'y': 2476, 'w': 66, 'h': 119},
    ]

    for output, target in zip(detects, targets):
        for key in target.keys():
            if key == 'l':
                assert output.get(key) == target.get(key)
            elif key == 'c':
                assert abs(output.get(key) - target.get(key)) < 1e-2
            else:
                assert abs(output.get(key) - target.get(key)) < 3



# -*- coding: utf-8 -*-
import os

import numpy as np
import PIL
import torch
import torchvision
import utool as ut

BATCH_SIZE = int(os.getenv('WIC_BATCH_SIZE', 256))
INPUT_SIZE = 224


class ImageFilePathList(torch.utils.data.Dataset):
    def __init__(self, filepaths, targets=None, transform=None, target_transform=None):
        from torchvision.datasets.folder import default_loader

        self.targets = targets is not None

        args = (filepaths, targets) if self.targets else (filepaths,)
        self.samples = list(zip(*args))

        if self.targets:
            self.classes = sorted(set(ut.take_column(self.samples, 1)))
            self.class_to_idx = {self.classes[i]: i for i in range(len(self.classes))}
        else:
            self.classes, self.class_to_idx = None, None

        self.loader = default_loader
        self.transform = transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        sample = self.samples[index]

        if self.targets:
            path, target = sample
        else:
            path = sample[0]
            target = None

        sample = self.loader(path)

        if self.transform is not None:
            sample = self.transform(sample)

        if self.target_transform is not None:
            target = self.target_transform(target)

        result = (sample, target) if self.targets else (sample,)

        return result

    def __len__(self):
        return len(self.samples)

    def __repr__(self):
        fmt_str = 'Dataset ' + self.__class__.__name__ + '\n'
        fmt_str += '    Number of samples: {}\n'.format(self.__len__())
        tmp = '    Transforms (if any): '
        fmt_str += '{}{}\n'.format(
            tmp, self.transform.__repr__().replace('\n', '\n' + ' ' * len(tmp))
        )
        tmp = '    Target Transforms (if any): '
        fmt_str += '{}{}'.format(
            tmp, self.target_transform.__repr__().replace('\n', '\n' + ' ' * len(tmp))
        )
        return fmt_str


class Augmentations(object):
    def __call__(self, img):
        img = np.array(img)
        return self.aug.augment_image(img)


class TestAugmentations(Augmentations):
    def __init__(self, **kwargs):
        from imgaug import augmenters as iaa

        self.aug = iaa.Sequential([iaa.Resize((INPUT_SIZE, INPUT_SIZE))])


def _init_transforms(**kwargs):
    transform = torchvision.transforms.Compose(
        [
            TestAugmentations(**kwargs),
            torchvision.transforms.Lambda(PIL.Image.fromarray),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(
                [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
            ),
        ]
    )
    return transform



# -*- coding: utf-8 -*-
#
#   Image and annotations preprocessing for lightnet networks
#   The image transformations work with both Pillow and OpenCV images
#   The annotation transformations work with brambox.annotations.Annotation objects
#   Copyright EAVISE
#
import collections
import logging

import numpy as np
from PIL import Image, ImageOps

from scoutbot.loc.transforms.util import BaseMultiTransform

log = logging.getLogger(__name__)

try:
    import cv2
except ImportError:
    log.warn('OpenCV is not installed and cannot be used')
    cv2 = None

__all__ = ['Letterbox']


class Letterbox(BaseMultiTransform):
    """Transform images and annotations to the right network dimensions.

    Args:
        dimension (tuple, optional): Default size for the letterboxing, expressed as a (width, height) tuple; Default **None**
        dataset (lightnet.data.Dataset, optional): Dataset that uses this transform; Default **None**

    Note:
        Create 1 Letterbox object and use it for both image and annotation transforms.
        This object will save data from the image transform and use that on the annotation transform.
    """

    def __init__(self, dimension=None, dataset=None):
        super().__init__(dimension=dimension, dataset=dataset)
        if self.dimension is None and self.dataset is None:
            raise ValueError(
                'This transform either requires a dimension or a dataset to infer the dimension'
            )

        self.pad = None
        self.scale = None
        self.fill_color = 127

    def __call__(self, data):
        if data is None:
            return None
        elif isinstance(data, collections.abc.Sequence):
            return self._tf_anno(data)
        elif isinstance(data, Image.Image):
            return self._tf_pil(data)
        elif isinstance(data, np.ndarray):
            return self._tf_cv(data)
        else:
            log.error(
                f'Letterbox only works with <brambox annotation lists>, <PIL images> or <OpenCV images> [{type(data)}]'
            )
            return data

    def _tf_pil(self, img):
        """Letterbox an image to fit in the network"""
        if self.dataset is not None:
            net_w, net_h = self.dataset.input_dim
        else:
            net_w, net_h = self.dimension
        im_w, im_h = img.size

        if im_w == net_w and im_h == net_h:
            self.scale = None
            self.pad = None
            return img

        # Rescaling
        if im_w / net_w >= im_h / net_h:
            self.scale = net_w / im_w
        else:
            self.scale = net_h / im_h
        if self.scale != 1:
            bands = img.split()
            bands = [
                b.resize((int(self.scale * im_w), int(self.scale * im_h))) for b in bands
            ]
            img = Image.merge(img.mode, bands)
            im_w, im_h = img.size

        if im_w == net_w and im_h == net_h:
            self.pad = None
            return img

        # Padding
        img_np = np.array(img)
        channels = img_np.shape[2] if len(img_np.shape) > 2 else 1
        pad_w = (net_w - im_w) / 2
        pad_h = (net_h - im_h) / 2
        self.pad = (int(pad_w), int(pad_h), int(pad_w + 0.5), int(pad_h + 0.5))
        img = ImageOps.expand(img, border=self.pad, fill=(self.fill_color,) * channels)
        return img

    def _tf_cv(self, img):
        """Letterbox and image to fit in the network"""
        if self.dataset is not None:
            net_w, net_h = self.dataset.input_dim
        else:
            net_w, net_h = self.dimension
        im_h, im_w = img.shape[:2]

        if im_w == net_w and im_h == net_h:
            self.scale = None
            self.pad = None
            return img

        # Rescaling
        if im_w / net_w >= im_h / net_h:
            self.scale = net_w / im_w
        else:
            self.scale = net_h / im_h
        if self.scale != 1:
            img = cv2.resize(
                img, None, fx=self.scale, fy=self.scale, interpolation=cv2.INTER_CUBIC
            )
            im_h, im_w = img.shape[:2]

        if im_w == net_w and im_h == net_h:
            self.pad = None
            return img

        # Padding
        # channels = img.shape[2] if len(img.shape) > 2 else 1
        pad_w = (net_w - im_w) / 2
        pad_h = (net_h - im_h) / 2
        self.pad = (int(pad_w), int(pad_h), int(pad_w + 0.5), int(pad_h + 0.5))
        img = cv2.copyMakeBorder(
            img,
            self.pad[1],
            self.pad[3],
            self.pad[0],
            self.pad[2],
            cv2.BORDER_CONSTANT,
            value=self.fill_color,
        )
        return img

    def _tf_anno(self, annos):
        """Change coordinates of an annotation, according to the previous letterboxing"""
        for anno in annos:
            if self.scale is not None:
                anno.x_top_left *= self.scale
                anno.y_top_left *= self.scale
                anno.width *= self.scale
                anno.height *= self.scale
            if self.pad is not None:
                anno.x_top_left += self.pad[0]
                anno.y_top_left += self.pad[1]
        return annos



# -*- coding: utf-8 -*-
#
#   Copyright EAVISE
#

from enum import Enum

__all__ = ['Box', 'ParserType', 'Parser']


class Box:
    """This is a generic bounding box representation.
    This class provides some base functionality to both annotations and detections.

    Attributes:
        class_label (string): class string label; Default **''**
        object_id (int): Object identifier for reid purposes; Default **None**
        x_top_left (Number): X pixel coordinate of the top left corner of the bounding box; Default **0.0**
        y_top_left (Number): Y pixel coordinate of the top left corner of the bounding box; Default **0.0**
        width (Number): Width of the bounding box in pixels; Default **0.0**
        height (Number): Height of the bounding box in pixels; Default **0.0**
    """

    def __init__(self):
        self.class_label = ''  # class string label
        self.object_id = None  # object identifier
        self.x_top_left = 0.0  # x pixel coordinate top left of the box
        self.y_top_left = 0.0  # y pixel coordinate top left of the box
        self.width = 0.0  # width of the box in pixels
        self.height = 0.0  # height of the box in pixels

    @classmethod
    def create(cls, obj=None):
        """Create a bounding box from a string or other detection object.

        Args:
            obj (Box or string, optional): Bounding box object to copy attributes from or string to deserialize
        """
        instance = cls()

        if obj is None:
            return instance

        if isinstance(obj, str):
            instance.deserialize(obj)
        elif isinstance(obj, Box):
            instance.class_label = obj.class_label
            instance.object_id = obj.object_id
            instance.x_top_left = obj.x_top_left
            instance.y_top_left = obj.y_top_left
            instance.width = obj.width
            instance.height = obj.height
        else:
            raise TypeError(
                'Object is not of type Box or not a string [obj.__class__.__name__]'
            )

        return instance

    def __eq__(self, other):
        # TODO: refactor -> use almost equal for floats
        return self.__dict__ == other.__dict__

    def serialize(self):
        """abstract serializer, implement in derived classes."""
        raise NotImplementedError

    def deserialize(self, string):
        """abstract parser, implement in derived classes."""
        raise NotImplementedError


class ParserType(Enum):
    """Enum for differentiating between different parser types."""

    UNDEFINED = 0  #: Undefined parsertype. Do not use this!
    SINGLE_FILE = 1  #: One single file contains all annotations
    MULTI_FILE = 2  #: One annotation file per image


class Parser:
    """This is a Generic parser class.

    Args:
        kwargs (optional): Derived parsers should use keyword arguments to get any information they need upon initialisation.
    """

    parser_type = (
        ParserType.UNDEFINED
    )  #: Type of parser. Derived classes should set the correct value.
    box_type = Box  #: Type of bounding box this parser parses or generates. Derived classes should set the correct type.
    extension = '.txt'  #: Extension of the files this parser parses or creates. Derived classes should set the correct extension.
    read_mode = 'r'  #: Reading mode this parser uses when it parses a file. Derived classes should set the correct mode.
    write_mode = 'w'  #: Writing mode this parser uses when it generates a file. Derived classes should set the correct mode.

    def __init__(self, **kwargs):
        pass

    def serialize(self, box):
        """Serialization function that can be overloaded in the derived class.
        The default serializer will call the serialize function of the bounding boxes and join them with a newline.

        Args:
            box: Bounding box objects

        Returns:
            string: Serialized bounding boxes

        Note:
            The format of the box parameter depends on the type of parser. |br|
            If it is a :any:`brambox.boxes.ParserType.SINGLE_FILE`, the box parameter should be a dictionary ``{"image_id": [box, box, ...], ...}``. |br|
            If it is a :any:`brambox.boxes.ParserType.MULTI_FILE`, the box parameter should be a list ``[box, box, ...]``.
        """
        if self.parser_type != ParserType.MULTI_FILE:
            raise TypeError(
                'The default implementation of serialize only works with MULTI_FILE'
            )

        result = ''
        for b in box:
            new_box = self.box_type.create(b)
            result += new_box.serialize() + '\n'

        return result

    def deserialize(self, string):
        """Deserialization function that can be overloaded in the derived class.
        The default deserialize will create new ``box_type`` objects and call the deserialize function of these objects with every line of the input string.

        Args:
            string (string): Input string to deserialize

        Returns:
            box: Bounding box objects

        Note:
            The format of the box return value depends on the type of parser. |br|
            If it is a :any:`brambox.boxes.ParserType.SINGLE_FILE`, the return value should be a dictionary ``{"image_id": [box, box, ...], ...}``. |br|
            If it is a :any:`brambox.boxes.ParserType.MULTI_FILE`, the return value should be a list ``[box, box, ...]``.
        """
        if self.parser_type != ParserType.MULTI_FILE:
            raise TypeError(
                'The default implementation of deserialize only works with MULTI_FILE'
            )

        result = []
        for line in string.splitlines():
            result += [self.box_type.create(line)]

        return result



# -*- coding: utf-8 -*-
from os.path import abspath, exists, join

import onnx


def test_loc_onnx_load():
    from scoutbot.loc import fetch

    onnx_model = fetch()
    model = onnx.load(onnx_model)
    assert exists(onnx_model)

    onnx.checker.check_model(model)

    graph = onnx.helper.printable_graph(model.graph)
    assert graph.count('\n') == 107


def test_loc_onnx_pipeline():
    from scoutbot.loc import INPUT_SIZE, post, pre, predict

    inputs = [
        abspath(join('examples', '0d01a14e-311d-e153-356f-8431b6996b84.true.jpg')),
    ]

    assert exists(inputs[0])

    data, sizes = pre(inputs)

    assert len(data) == 1
    assert len(data[0]) == 3
    assert len(data[0][0]) == INPUT_SIZE[0]
    assert len(data[0][0][0]) == INPUT_SIZE[1]
    assert sizes == [(256, 256)]

    preds = predict(data)

    assert len(preds) == 1
    assert len(preds[0]) == 30

    outputs = post(preds, sizes)

    assert len(outputs) == 1
    assert len(outputs[0]) == 5

    # fmt: off
    targets = [
        {
            'l': 'elephant_savanna',
            'x': 206.00893930,
            'y': 189.09138371,
            'w':  53.78145658,
            'h':  66.46106896,
            'c':   0.77065581,
        },
        {
            'l': 'elephant_savanna',
            'x': 216.61065204,
            'y': 193.30525090,
            'w':  42.83404541,
            'h':  62.44728440,
            'c':   0.61152166,
        },
        {
            'l': 'elephant_savanna',
            'x':  51.61210749,
            'y': 235.37819260,
            'w':  79.69709660,
            'h':  17.41258826,
            'c':   0.50862342,
        },
        {
            'l': 'elephant_savanna',
            'x':  57.47630427,
            'y': 236.92587515,
            'w':  94.69935960,
            'h':  16.03246718,
            'c':   0.44841822,
        },
        {
            'l': 'elephant_savanna',
            'x':  37.07233605,
            'y': 230.39122596,
            'w': 105.40560208,
            'h':  24.81017362,
            'c':   0.44012001,
        },
    ]
    # fmt: on

    for output, target in zip(outputs[0], targets):
        for key in target.keys():
            if key == 'l':
                assert output.get(key) == target.get(key)
            elif key == 'c':
                assert abs(output.get(key) - target.get(key)) < 1e-2
            else:
                assert abs(output.get(key) - target.get(key)) < 3



# -*- coding: utf-8 -*-
from os.path import abspath, exists, join, relpath

import cv2
import numpy as np
import utool as ut


def test_tile_grid():
    from scoutbot.tile import TILE_WIDTH, tile_grid

    img_filepath = abspath(join('examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41.jpg'))
    assert exists(img_filepath)
    img = cv2.imread(img_filepath)

    assert img.shape == (4000, 6016, 3)

    grid1 = tile_grid(img.shape)
    grid2 = tile_grid(img.shape, offset=TILE_WIDTH // 2, borders=False)
    grid = grid1 + grid2

    assert len(grid1) == 682
    assert len(grid2) == 570
    assert len(set(map(str, grid))) == len(grid)

    assert grid1[0] == {'x': 0, 'y': 0, 'w': 256, 'h': 256, 'b': True}
    assert grid2[0] == {'x': 128, 'y': 176, 'w': 256, 'h': 256, 'b': False}


def test_tile_filepath():
    from scoutbot.tile import tile_filepath

    img_filepath = abspath(join('examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41.jpg'))
    assert exists(img_filepath)

    grid = {'x': 0, 'y': 0, 'w': 256, 'h': 256, 'b': True}

    filepath = tile_filepath(img_filepath, grid)
    filepath = relpath(filepath)
    assert filepath == join(
        'examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41_x_0_y_0_w_256_h_256.jpg'
    )

    if exists(filepath):
        ut.delete(filepath, verbose=False)


def test_tile_write():
    from scoutbot.tile import tile_filepath, tile_write

    img_filepath = abspath(join('examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41.jpg'))
    assert exists(img_filepath)
    img = cv2.imread(img_filepath)

    grid = {'x': 0, 'y': 0, 'w': 256, 'h': 256, 'b': True}
    filepath = tile_filepath(img_filepath, grid, ext='.png')  # Use lossless here
    assert tile_write(img, grid, filepath)

    assert exists(filepath)
    tile_img = cv2.imread(filepath)

    x0 = grid.get('x')
    y0 = grid.get('y')
    w = grid.get('w')
    h = grid.get('h')
    y1 = y0 + h
    x1 = x0 + w
    tile = img[y0:y1, x0:x1]

    assert tile_img.shape == (256, 256, 3)
    assert tile.shape == (256, 256, 3)

    tile = tile.astype(np.float32)
    tile_img = tile_img.astype(np.float32)

    assert np.all(np.abs(tile - tile_img) <= 0)
    ut.delete(filepath, verbose=False)


def test_tile_compute():
    from scoutbot.tile import compute

    img_filepath = abspath(join('examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41.jpg'))
    shape, grids, filepaths = compute(img_filepath)

    assert len(filepaths) == 1252
    for filepath in filepaths:
        assert exists(filepath)
        ut.delete(filepath, verbose=False)



# -*- coding: utf-8 -*-
from os.path import abspath, exists, join

import onnx


def test_wic_onnx_load():
    from scoutbot.wic import fetch

    onnx_model = fetch()
    model = onnx.load(onnx_model)
    assert exists(onnx_model)

    onnx.checker.check_model(model)

    graph = onnx.helper.printable_graph(model.graph)
    assert graph.count('\n') == 1334


def test_wic_onnx_pipeline():
    from scoutbot.wic import INPUT_SIZE, ONNX_CLASSES, post, pre, predict

    inputs = [
        abspath(join('examples', '1e8372e4-357d-26e6-d7fd-0e0ae402463a.true.jpg')),
    ]

    assert exists(inputs[0])

    data = pre(inputs)

    assert len(data) == 1
    assert len(data[0]) == 3
    assert len(data[0][0]) == INPUT_SIZE
    assert len(data[0][0][0]) == INPUT_SIZE

    preds = predict(data)

    assert len(preds) == 1
    assert len(preds[0]) == 2
    assert preds[0][1] > preds[0][0]
    assert abs(preds[0][0] - 0.00001503) < 1e-4
    assert abs(preds[0][1] - 0.99998497) < 1e-4

    outputs = post(preds)

    assert len(outputs) == 1
    output = outputs[0]
    assert output.keys() == set(ONNX_CLASSES)
    assert output['positive'] > output['negative']
    assert abs(output['negative'] - 0.00001503) < 1e-4
    assert abs(output['positive'] - 0.99998497) < 1e-4



# -*- coding: utf-8 -*-
from os.path import abspath, join

import scoutbot


def test_fetch():
    scoutbot.fetch(pull=False)
    scoutbot.fetch(pull=True)


def test_pipeline():
    img_filepath = abspath(join('examples', '1be4d40a-6fd0-42ce-da6c-294e45781f41.jpg'))

    detects = scoutbot.pipeline(img_filepath)
    assert len(detects) == 3

    targets = [
        {'l': 'elephant_savanna', 'c': 0.9299, 'x': 4597, 'y': 2322, 'w': 72, 'h': 149},
        {'l': 'elephant_savanna', 'c': 0.8739, 'x': 4865, 'y': 2422, 'w': 97, 'h': 109},
        {'l': 'elephant_savanna', 'c': 0.7115, 'x': 4806, 'y': 2476, 'w': 66, 'h': 119},
    ]

    for output, target in zip(detects, targets):
        for key in target.keys():
            if key == 'l':
                assert output.get(key) == target.get(key)
            elif key == 'c':
                assert abs(output.get(key) - target.get(key)) < 1e-2
            else:
                assert abs(output.get(key) - target.get(key)) < 3



# -*- coding: utf-8 -*-
#
#   Copyright EAVISE
#

# from enum import Enum

from scoutbot.loc.transforms import box as b

__all__ = ['Detection', 'ParserType', 'Parser']


class Detection(b.Box):
    """This is a generic detection class that provides some base functionality all detections need.
    It builds upon :class:`~brambox.boxes.box.Box`.

    Attributes:
        confidence (Number): confidence score between 0-1 for that detection; Default **0.0**
    """

    def __init__(self):
        """x_top_left,y_top_left,width,height are in pixel coordinates"""
        super(Detection, self).__init__()
        self.confidence = 0.0  # Confidence score between 0-1

    @classmethod
    def create(cls, obj=None):
        """Create a detection from a string or other box object.

        Args:
            obj (Box or string, optional): Bounding box object to copy attributes from or string to deserialize

        Note:
            The obj can be both an :class:`~brambox.boxes.annotations.Annotation` or a :class:`~brambox.boxes.detections.Detection`.
            For Detections the confidence score is copied over, for Annotations it is set to 1.
        """
        from scoutbot.loc.transforms.annotations import annotation as anno

        instance = super(Detection, cls).create(obj)

        if obj is None:
            return instance

        if isinstance(obj, Detection):
            instance.confidence = obj.confidence
        elif isinstance(obj, anno.Annotation):
            instance.confidence = 1.0

        return instance

    def __repr__(self):
        """Unambiguous representation"""
        string = f'{self.__class__.__name__} ' + '{'
        string += f'class_label = {self.class_label}, '
        string += f'object_id = {self.object_id}, '
        string += f'x = {self.x_top_left}, '
        string += f'y = {self.y_top_left}, '
        string += f'w = {self.width}, '
        string += f'h = {self.height}, '
        string += f'confidence = {self.confidence}'
        return string + '}'

    def __str__(self):
        """Pretty print"""
        string = 'Detection {'
        string += f'\'{self.class_label}\'{"" if self.object_id is None else " "+str(self.object_id)}, '
        string += f'[{int(self.x_top_left)}, {int(self.y_top_left)}, {int(self.width)}, {int(self.height)}]'
        string += f', {round(self.confidence*100, 2)}%'
        return string + '}'

    def serialize(self, return_dict=False):
        import json

        serialize_list = [
            self.class_label,
            self.object_id,
            self.x_top_left,
            self.y_top_left,
            self.width,
            self.height,
            self.confidence,
        ]
        if return_dict:
            return serialize_list
        else:
            serialize_str = json.dumps(serialize_list)
            return serialize_str

    def deserialize(self, serialize_str, input_dict=False):
        import json

        if input_dict:
            assert isinstance(serialize_str, dict)
            serialize_list = serialize_str
        else:
            serialize_list = json.loads(serialize_str)
        self.class_label = serialize_list[0]
        self.object_id = serialize_list[1]
        self.x_top_left = serialize_list[2]
        self.y_top_left = serialize_list[3]
        self.width = serialize_list[4]
        self.height = serialize_list[5]
        self.confidence = serialize_list[6]
        return True


ParserType = b.ParserType


class Parser(b.Parser):
    """Generic parser class"""

    box_type = Detection  # Derived classes should set the correct box_type



# -*- coding: utf-8 -*-
'''
Scoutbot utilities file for common and handy functions.
'''
import logging
from logging.handlers import TimedRotatingFileHandler

DAYS = 21


def init_logging():
    """
    Setup Python's built in logging functionality with on-disk logging, and prettier logging with Rich
    """
    # Import Rich
    import rich
    from rich.logging import RichHandler
    from rich.style import Style
    from rich.theme import Theme

    name = 'scoutbot'

    # Setup placeholder for logging handlers
    handlers = []

    # Configuration arguments for console, handlers, and logging
    console_kwargs = {
        'theme': Theme(
            {
                'logging.keyword': Style(bold=True, color='yellow'),
                'logging.level.notset': Style(dim=True),
                'logging.level.debug': Style(color='cyan'),
                'logging.level.info': Style(color='green'),
                'logging.level.warning': Style(color='yellow'),
                'logging.level.error': Style(color='red', bold=True),
                'logging.level.critical': Style(color='red', bold=True, reverse=True),
                'log.time': Style(color='white'),
            }
        )
    }
    handler_kwargs = {
        'rich_tracebacks': True,
        'tracebacks_show_locals': True,
    }
    logging_kwargs = {
        'level': logging.INFO,
        'format': '[%(name)s] %(message)s',
        'datefmt': '[%X]',
    }

    # Add file-baesd log handler
    handlers.append(
        TimedRotatingFileHandler(
            filename=f'{name}.log',
            when='midnight',
            backupCount=DAYS,
        ),
    )

    # Add rich (fancy logging) log handler
    rich.reconfigure(**console_kwargs)
    handlers.append(RichHandler(**handler_kwargs))

    # Setup global logger with the handlers and set the default level to INFO
    logging.basicConfig(handlers=handlers, **logging_kwargs)
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    log = logging.getLogger(name)

    return log
